/*
 *  Copyright (C) Chaperon. All rights reserved.
 *  -------------------------------------------------------------------------
 *  This software is published under the terms of the Apache Software License
 *  version 1.1, a copy of which has been included  with this distribution in
 *  the LICENSE file.
 */

package net.sourceforge.chaperon.grammar.production;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import net.sourceforge.chaperon.grammar.Grammar;
import net.sourceforge.chaperon.grammar.symbol.Symbol;
import net.sourceforge.chaperon.grammar.symbol.SymbolList;
import net.sourceforge.chaperon.helpers.IntegerList;

/**
 * This class presents a list of productions
 *
 * @author Stephan Michels
 * @version $Revision: 1.5 $ $Date: 2002/05/06 16:22:56 $
 */
public class ProductionList implements Serializable, Cloneable
{

	private Grammar _owner = null;
	private Vector _list = new Vector();

	/**
	 * Creates a empty list
	 */
	public ProductionList()
	{
	}

	/**
	 *
	 *
	 * @param owner
	 */
	public void setOwner(Grammar owner)
	{
		if (_owner != null)
			throw new IllegalStateException("ProductionList has already a owner");
		_owner = owner;
	}

	/**
	 *
	 *
	 * @return
	 */
	public Grammar getOwner()
	{
		return _owner;
	}

	/**
	 * Add a production to this list
	 *
	 * @param production Production, which should added
	 *
	 * @return Index of the production in this list
	 */
	public int addProduction(Production production)
	{
    if (production==null)
      throw new NullPointerException();

    production.setOwner(this);

  	_list.addElement(production);
	  return _list.size() - 1;
	}

	/**
	 * Add the content of an other list to this list
	 *
	 * @param list List of productions
	 */
	public void addProduction(ProductionList list)
	{
		for (int i = 0; i < list.getProductionCount(); i++)
    {
      try
      {
        addProduction((Production)list.getProduction(i).clone());
      } catch (CloneNotSupportedException cnse) 
      {
        throw new IllegalArgumentException("Could not clone token:"+cnse.getMessage());
      }
    }
	}

	/**
	 * Removes a production by an index from this list
	 *
	 * @param index Index of the production, which should be removed
	 */
	public void removeProduction(int index)
	{
		_list.removeElementAt(index);
	}

	/**
	 * Set a production by an index
	 *
	 * @param index The index, at which the production be inserted
	 * @param production Production
	 */
	public void setProduction(int index,
														Production production)
															throws IndexOutOfBoundsException
	{
		if ((index < 0) || (index > _list.size()))
			throw new IndexOutOfBoundsException();
		_list.setElementAt(production, index);
	}

	/**
	 * Return a production giving by an index
	 *
	 * @param index Index of the Production
	 *
	 * @return Production
	 */
	public Production getProduction(int index) throws IndexOutOfBoundsException
	{
		if ((index < 0) || (index > _list.size()))
			throw new IndexOutOfBoundsException();

		return (Production) _list.elementAt(index);
	}

	/**
	 * Returns all production for given non terminal symbol
	 *
	 * @param ntsymbol Non terminal symbol
	 *
	 * @return List of indices from found productions
	 */
	public IntegerList getProductionList(Symbol ntsymbol)
	{
		IntegerList list = new IntegerList();
		int i;

		for (i = 0; i < getProductionCount(); i++)
		{
			if (getProduction(i).getSymbol().equals(ntsymbol))
				list.add(i);
		}
		return list;
	}

	/**
	 * Returns the count of productions in the list
	 *
	 * @return Count of productions
	 */
	public int getProductionCount()
	{
		return _list.size();
	}

	/**
	 * Return the index of a production
	 *
	 * @param production Production
	 *
	 * @return Index of the Production
	 */
	public int indexOf(Production production)
	{
		for (int i = 0; i < _list.size(); i++)
			if (((Production) _list.elementAt(i)).equals(production))
				return i;
		return -1;
	}

	/**
	 * Return the index of the next production, which found
	 * by a non terminal symbol
	 *
	 * @param ntsymbol Non terminal symbol
	 *
	 * @return Production, which found
	 */
	public int indexOf(Symbol ntsymbol)
	{
		for (int i = 0; i < _list.size(); i++)
			if (((Production) _list.elementAt(i)).getSymbol().equals(ntsymbol))
				return i;
		return -1;
	}

	/**
	 * If the list contains a production
	 *
	 * @param production Production
	 *
	 * @return True, if the list contains the production
	 */
	public boolean contains(Production production)
	{
		return (indexOf(production) != -1);
	}

	/**
	 * If the list contains a production with a special symbol
	 *
	 * @param ntsymbol Non terminal symbol
	 *
	 * @return True, if the list contains a production with the symbol
	 */
	public boolean contains(Symbol ntsymbol)
	{
		return (indexOf(ntsymbol) != -1);
	}

	/**
	 * Removes all productions from this list
	 */
	public void removeAllProduction()
	{
		_list.removeAllElements();
	}

	/**
	 * Returns a enumeration of the productions
	 * in this list.
	 *
	 * @return Enumeration of the productions
	 */
	public Enumeration enumerateProduction()
	{
		return _list.elements();
	}

	/**
	 * Returns the content of this list
	 *
	 * @return Array of the tokens
	 */
	public Production[] getProduction()
	{
		int size = _list.size();
		Production[] mArray = new Production[size];

		for (int index = 0; index < size; index++)
			mArray[index] = (Production) _list.elementAt(index);
		return mArray;
	}

	/**
	 * Replace the content of this list by the content of an array
	 *
	 *
	 * @param productionArray
	 */
	public void setProduction(Production[] productionArray)
	{
		_list.removeAllElements();
		for (int i = 0; i < productionArray.length; i++)
			_list.addElement(productionArray[i]);
	}

	/**
	 * Return all used symbols in this list
	 *
	 * @return List of symbols
	 */
	public SymbolList getSymbols()
	{
		SymbolList list = new SymbolList(true);

		for (int i = 0; i < getProductionCount(); i++)
			list.addSymbolList(getProduction(i).getSymbols());
		return list;
	}

	/**
	 * Convert this list to a string
	 *
	 * @return Content of this list
	 */
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append("Produktions:\n");
		for (int i = 0; i < getProductionCount(); i++)
		{
			buffer.append(i + ".Production: ");
			buffer.append(getProduction(i).toString());
			buffer.append("\n");
		}

		buffer.append("\n");
		return buffer.toString();
	}

	/**
	 *
	 *
	 * @return
	 *
	 * @throws CloneNotSupportedException
	 */
	public Object clone() throws CloneNotSupportedException
	{
		ProductionList clone = new ProductionList();

		for (int i = 0; i < _list.size(); i++)
			clone.addProduction((Production) ((Production) _list.elementAt(i)).clone());

		return clone;
	}
}
