/*
 *  Copyright (C) Chaperon. All rights reserved.                               
 *  ------------------------------------------------------------------------- 
 *  This software is published under the terms of the Apache Software License 
 *  version 1.1, a copy of which has been included  with this distribution in 
 *  the LICENSE file.                                                         
 */

package net.sourceforge.chaperon.grammar.symbol;

import java.io.Serializable;
import java.util.Vector;

/**
 * This class represents a list of symbols
 *
 * @author Stephan Michels
 * @version CVS $Id: SymbolList.java,v 1.3 2002/05/06 16:22:56 benedikta Exp $
 */
public class SymbolList implements Serializable, Cloneable
{
  private int capacityIncrement = 100;

  private int elementCount = 0;
  private Symbol[] list = new Symbol[10];

  private boolean unit = false;

  /**
   * Creates a empty list
   */
  public SymbolList()
  {
  }

  /**
   * Creates a empty list
   *
   * @param unit If true, this list contains no multiple instances
   *             of the same symbol
   */
  public SymbolList(boolean unit)
  {
    this.unit = unit;
  }

  /**
   * Add a symbol to his list
   *
   * @param symbol Symbol, which should added
   *
   * @return Index of the symbol in this list
   */
  public int addSymbol(Symbol symbol)
  {
    if (symbol==null)
    {
      System.err.println("SymbolList.add:Symbol is null");
      return -1;
    }

    if (unit)
    {
      int index = indexOf(symbol);

      if (index == -1)
      {
        ensureCapacity(elementCount + 1);
        list[elementCount] = symbol;
        return elementCount++;
      }
      else
        return index;
    }

    ensureCapacity(elementCount + 1);
    list[elementCount] = symbol;
    return elementCount++;
  }

  /**
   * Add the symbols from an other list to this list
   *
   * @param list Lst of symbol
   */
  public void addSymbolList(SymbolList list)
  {
    for (int i = 0; i < list.getSymbolCount(); i++)
      addSymbol(list.getSymbol(i));
  }

  /**
   * Removes a symbol by an index from this list
   *
   * @param index Index of the symbol
   */
  public void removeSymbol(int index)
  {
    if (index >= elementCount)
      throw new ArrayIndexOutOfBoundsException(index);

    elementCount--;
    if (index < elementCount)
      System.arraycopy(list, index + 1, list, index, elementCount - index);
  }

  /**
   * Removes a symbol from this list
   *
   * @param symbol Symbol, which should be removed
   */
  public void removeSymbol(Symbol symbol)
  {
    for (int i = elementCount - 1; i >= 0; i--)
      if (symbol.equals(list[i]))
        removeSymbol(i);
  }

  /**
   * Set a symbol by an index
   *
   * @param index The index, at which the symbol be inserted
   * @param symbol Symbol
   */
  public void setSymbol(int index, Symbol symbol)
  {
    if (index >= elementCount)
      throw new ArrayIndexOutOfBoundsException(index);

    list[index] = symbol;
  }

  /**
   * Return a symbol giving by an index
   *
   * @param index Index of the symbol
   *
   * @return Symbol
   */
  public Symbol getSymbol(int index)
  {
    if ((index >= 0) && (index < elementCount))
      return list[index];
    return null;
  }

  /**
   * Returns a symbol from this list given by the name
   * of the symbol
   *
   * @param name Name of the symbol
   *
   * @return Symbol
   */
  public Symbol getSymbol(String name)
  {
    int index = indexOf(name);

    if (index >= 0)
      return list[index];
    return null;
  }

  /**
   * Returns the count of symbols in the list
   *
   * @return Count of symbol
   */
  public int getSymbolCount()
  {
    return elementCount;
  }

  /**
   * Return the index of a symbol
   *
   * @param symbol Symbol
   *
   * @return Index of this symbol
   */
  public int indexOf(Symbol symbol)
  {
    for (int i = 0; i < elementCount; i++)
      if (list[i].equals(symbol))
        return i;
    return -1;
  }

  /**
   * Return the index of a symbol, given by 
   * the name of the Symbol
   *
   * @param name Name of symbol
   *
   * @return Index of this symbol
   */
  public int indexOf(String name)
  {
    for (int i = 0; i < elementCount; i++)
      if (list[i].getName().equals(name))
        return i;
    return -1;
  }

  /**
   * If the list contains a symbol
   *
   * @param symbol Symbol
   *
   * @return True, if the list contains the symbol
   */
  public boolean contains(Symbol symbol)
  {
    return (indexOf(symbol) != -1);
  }

  /**
   * If the list contains a symbol, given by 
   * the name of the symbol
   *
   * @param name Name of the symbol
   *
   * @return True, if the list contains the symbol
   */
  public boolean contains(String name)
  {
    return (indexOf(name) != -1);
  }

  /**
   * Removes all symbols from this list
   */
  public void clear()
  {
    elementCount = 0;
  }

  /**
   * Return all used terminal symbols in this list
   *
   * @return List of terminal symbols
   */
  public SymbolList getTerminalSymbols()
  {
    SymbolList list = new SymbolList(unit);
    Symbol symbol;

    for (int i = 0; i < elementCount; i++)
      if ((symbol = this.list[i]).isTerminal())
        list.addSymbol(symbol);
    return list;
  }

  /**
   * Return all used non terminal symbols in this list
   *
   * @return List of non terminal symbols
   */
  public SymbolList getNonTerminalSymbols()
  {
    SymbolList list = new SymbolList(unit);
    Symbol symbol;

    for (int i = 0; i < elementCount; i++)
      if ((symbol = this.list[i]).isNonTerminal())
        list.addSymbol(symbol);
    return list;
  }

  /**
   * Convert this list to a string
   *
   * @return Content of this list
   */
  public String toString()
  {
    StringBuffer buffer = new StringBuffer();

    for (int i = 0; i < elementCount; i++)
    {
      buffer.append(list[i].toString());
      if (i < (elementCount - 1))
        buffer.append(" ");
    }
    return buffer.toString();
  }

  public Object clone() throws CloneNotSupportedException
  {
    SymbolList clone = new SymbolList();

    for(int i=0; i<getSymbolCount(); i++)
      clone.addSymbol(getSymbol(i));

    return clone;
  }

  /**
   * Ensure the capacity for adding values
   *
   * @param minCapacity
   */
  private void ensureCapacity(int minCapacity)
  {
    if (list.length >= minCapacity)
      return;

    int newCapacity = list.length + capacityIncrement;

    if (capacityIncrement <= 0)
      newCapacity = list.length * 2;

    Symbol[] newArray = new Symbol[Math.max(newCapacity, minCapacity)];

    System.arraycopy(list, 0, newArray, 0, list.length);
    list = newArray;
  }
}
